/*
Copyright (c) Dale Anson, 2004
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote products
derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package ise.dbconsole;

import java.io.*;
import java.sql.*;
import java.util.*;
import javax.sql.DataSource;

import ise.library.*;
import ise.library.ascii.*;

import jline.*;

/**
 * A console application to run database commands. User may enter queries or
 * several application commands. Applications commands are single line commands
 * that start with '\', for example '\q' will quit the application. For all
 * other user input, the first word of the input is taken as the command, the
 * rest of the input is taken as data for that command. For example, if the user
 * enters 'select * from mytable;', then 'select' is the command and '* from
 * mytable' is the data for that command. Such entries may be across several
 * lines, processing occurs when a line ends with a ';'. This uses the "Command"
 * design pattern. Commands are defined in commands.conf. All command names are
 * lowercase in command.conf, although the actual user input may be mixed case.
 * This means that "select", "SELECT", and "Select" will all use the same
 * command processor.
 *
 * @author    Dale Anson, Sep 2004
 * @version   $Revision: 18 $
 */
public final class DbConsole implements Constants {

    // keep track of the original System.in and System.out
    private final static InputStream _system_in = System.in;
    private final static PrintStream _system_out = System.out;

    // console input
    private InputStream in = null;

    // console output
    private PrintStream out = null;

    // ConsoleReader, from jline, a special reader for console input
    private static ConsoleReader consoleReader = null;

    private Thread commandLoop = null;
    private volatile boolean canLoop = true;

    // list of commands, key is a string naming the command, value is the
    // classname of the command processor
    private Properties command_lu;

    // storage for the Commands, key is the classname of the command, value is a
    // Command.
    private HashMap commands;

    private ArrayList saved_queries = new ArrayList( 20 );

    // data access object
    private DbConsoleDAO dao;

    private volatile boolean canSpin = false;   // spinner thread control
    private boolean doSpin = true;   // permission to spin
    private Thread spinner;
    private String working = "Working...";
    private StopWatch stopwatch = new StopWatch();
    private CommandThread runner = null;

    private boolean echo = false;

    public static String prompt = "> ";
    public static String msg_prompt = "-> ";

    private static String lastCommand = null;

    private static DbConsole instance = null;

    /** Constructor for DbConsole */
    private DbConsole() {
        MemoryMonitor mm = new MemoryMonitor( 5 );
        mm.addObserver( new Observer() {
                    public void update( Observable o, Object arg ) {
                        println( "Memory Usage: " + arg.toString() );
                    }
                }
                      );
    }

    /**
     * Gets the singleton instance of DbConsole
     *
     * @return   The instance
     */
    public static DbConsole getInstance() {
        if ( instance == null ) {
            instance = new DbConsole();
            instance.init();
        }
        return instance;
    }

    /**
     * Gets the DAO
     *
     * @return   The data access object
     */
    public DbConsoleDAO getDAO() {
        return dao;
    }

    /**
     * Sets the DAO
     *
     * @param dao  The new DAO
     */
    public void setDAO( DbConsoleDAO dao ) {
        this.dao = dao;
    }

    /**
     * The main command loop. The app sits here, waiting for user input, and
     * dispatches that input to the appropriate command.
     */
    private void createCommandLoop() {
        commandLoop =
            new Thread() {
                public void run() {
                    StringBuffer line_buffer = new StringBuffer();
                    StringBuffer command_buffer = new StringBuffer();
                    String last_command = null;
                    try {
                        do {
                            // I could just do a readLine here, but I have plans
                            // for processing on a character basis.
                            int c = in.read();
                            if ( c != -1 ) {
                                line_buffer.append( ( char ) c );
                            }
                            else {
                                line_buffer.append( LS );
                            }

                            /* might need to change this... LS may not be correct
                            depending on how the app is ran.  Some cygwin shells
                            append the Unix line ender rather than the DOS line
                            ender, so this next line doesn't work correctly. */
                            if ( !line_buffer.toString().endsWith( LS ) )
                                continue;

                            String line = line_buffer.toString().trim();
                            if ( line.length() == 0 )
                                continue;
                            if ( echo )
                                println( line );

                            line_buffer = new StringBuffer();

                            // there are a few special commands (let's keep this list small!)
                            // \r is special, it clears the buffer. Note this is a
                            // backslash followed by an r, not the carriage return character.
                            if ( line.trim().equals( "\\r" ) ) {
                                command_buffer = new StringBuffer();
                                println( msg_prompt + Bundle.BUNDLE.getString( "bufferCleared" ) );
                                continue;
                            }
                            // \rr is special, it reruns the last command
                            if ( line.trim().equals( "\\rr" ) ) {
                                if ( last_command != null ) {
                                    runner = new CommandThread( last_command );
                                    runner.start();
                                }
                                continue;
                            }
                            // \s is special, it stops the command that is currently executing
                            // -- no it doesn't, this is broke and needs fixed...
                            if ( line.trim().equals( "\\s" ) ) {
                                if ( runner != null && runner.isAlive() ) {
                                    Thread r = runner;
                                    runner = null;
                                    r.interrupt();
                                }
                                command_buffer = new StringBuffer();
                                println( msg_prompt );
                                continue;
                            }
                            // \o is special, it redirects output to a file.  In this case,
                            // turn on or off permission for the spinner to work so spinner
                            // output is not recorded in the file.
                            if ( line.trim().startsWith( "\\o" ) ) {
                                String[] args = line.split( "[ ]" );
                                doSpin = args.length == 1;
                            }
                            // ignore sql comments
                            // this needs more work since C-style comments are allowed
                            // to span multiple lines
                            if ( line.trim().startsWith( "/*" ) || line.trim().startsWith( "--" ) ) {
                                continue;
                            }

                            // application commands start with \, like \q for quit.
                            // Don't run these in a CommandThread.
                            if ( command_buffer.length() == 0 && line.startsWith( "\\" ) ) {
                                DbConsoleDAO.CAN_RUN = true;
                                executeCommand( line );
                                last_command = line;
                                command_buffer = new StringBuffer();
                            }
                            else {
                                command_buffer.append( line );
                                if ( line.endsWith( ";" ) ) {
                                    // run the command in a CommandThread.  Using a CommandThread
                                    // lets the \s command stop the thread and command if
                                    // necessary.
                                    command_buffer.append( " " );
                                    String command = command_buffer.toString();
                                    last_command = command;
                                    saved_queries.add(command);
                                    runner = new CommandThread( command );
                                    runner.start();
                                    try {
                                        /// might need to wait, there seems to be problems when
                                        /// running too many commands at once
                                        runner.join();
                                    }
                                    catch ( Exception ignored ) {
                                    }
                                    command_buffer = new StringBuffer();
                                }
                                else {
                                    command_buffer.append( " " );
                                    print( prompt );
                                }
                            }
                            if ( c == -1 ) {
                                resetIn();
                                return ;
                            }
                        }
                        while ( canLoop );
                    }
                    catch ( IOException e ) {
                        e.printStackTrace();
                        System.out.println( Bundle.BUNDLE.getString( "looksLikeThereWasAProblem" ) );
                        System.exit( 1 );
                    }
                }
            };
    }

    /** Initialize the application. */
    private void init() {
        // load stored user settings
        loadUserSettings();

        // load the application resource bundle, need the message strings for
        // the rest of the initialization
        loadResourceBundle();

        // load the specific commands
        loadCommands();

        // set up the IO
        setupIO();

        // set up the application prompts and show the intro
        setupApp();

        // show the database menu and set up the DAO
        executeCommand( "\\db" );

        startCommandLoop();

    }

    private void loadUserSettings() {
        boolean on = DbConsoleUtil.getShowRowColumn();
        System.setProperty( SHOW_ROW_COLUMN, on ? YES : NO );
        on = DbConsoleUtil.getShowColor();
        System.setProperty( SHOW_COLOR, on ? YES : NO );
        String color = DbConsoleUtil.getColor();
        System.setProperty( COLOR, color );
        int box_width = DbConsoleUtil.getBoxWidth();
        MessageBox.setMaxWidth( box_width );
        int saved_query_size = DbConsoleUtil.getQueryHistorySize();
    }

    /** Load the application messages. */
    private void loadResourceBundle() {
        try {
            InputStream is = getClass().getClassLoader().getResourceAsStream( "ise/dbconsole/bundle.props" );
            new Bundle( is );
        }
        catch ( Exception e ) {
            System.out.println( Bundle.BUNDLE.getString( "bundlePropsWarning" ) );
            System.exit( 1 );
        }
    }

    /**
     * Load the Commands from the configuration file. Each Command is
     * instantiated and stored in a hash map for later use/reuse. Load the
     * resource bundle.
     */
    private void loadCommands() {
        // load the commands
        try {
            InputStream is = getClass().getClassLoader().getResourceAsStream( "ise/dbconsole/commands.conf" );
            command_lu = new Properties();
            command_lu.load( is );
            commands = new HashMap();
            Enumeration en = command_lu.propertyNames();
            while ( en.hasMoreElements() ) {
                String command_name = ( String ) en.nextElement();
                String classname = command_lu.getProperty( command_name );
            }
        }
        catch ( Exception e ) {
            // show it, but ignore it for now
            e.printStackTrace();
            System.out.println( Bundle.BUNDLE.getString( "commandsConfWarning" ) );
            System.exit( 1 );
        }
    }

    /** Gets the command prompts, sets up IO, and shows the intro. */
    private void setupApp() {
        // load the app command prompt
        try {
            prompt = Bundle.BUNDLE.getString( "commandPrompt" ) + " ";
            msg_prompt = Bundle.BUNDLE.getString( "messagePrompt" ) + " ";
            working = Bundle.BUNDLE.getString( "working" );
        }
        catch ( Exception e ) {
            // ignored, just use defaults.
        }

        // show the intro
        executeCommand( "\\intro" );
        println();
    }

    /** Set up the IO streams. */
    private void setupIO() {
        in = _system_in;
        out = _system_out;

        // set up console input
        //resetIn();

        // set up console output
        //resetOut();
        try {
            consoleReader = new ConsoleReader();
            consoleReader.setUseHistory( true );
            in = new ConsoleReaderInputStream( consoleReader );
            System.setIn( in );
        }
        catch ( Exception e ) {
            e.printStackTrace();
        }

        println();
    }

    public Collection getSavedQueries() {
        return saved_queries;
    }

    public void setSavedQueries( ArrayList s ) {
        saved_queries = s;
    }

    public void clearSavedQueries() {
        saved_queries.clear();
    }

    public void addToCommandHistory( String command ) {
        consoleReader.getHistory().addToHistory( command );
    }

    public static String getLastCommand() {
        return lastCommand;
    }

    /** prints a blank line to the console */
    private void println() {
        try {
            out.print( LS );
            out.print( prompt );   // our command prompt
            out.flush();
        }
        catch ( Exception e ) {
            e.printStackTrace();
        }
    }

    /**
     * Prints the given string to the console followed by a line feed.
     *
     * @param s
     */
    private void println( String s ) {
        try {
            out.print( s );
            println();
        }
        catch ( Exception e ) {
            e.printStackTrace();
        }
    }

    /**
     * Prints the given string to the console .
     *
     * @param s
     */
    public void print( String s ) {
        try {
            out.print( s );
            out.flush();
        }
        catch ( Exception e ) {
            e.printStackTrace();
        }
    }

    /**
     * Removes count characters from the end of the current line in the display.
     *
     * @param count  the number of characters to remove.
     */
    private void erase( int count ) {
        try {
            for ( int i = 0; i < count; i++ )
                out.print( "\b" );
            for ( int i = 0; i < count; i++ )
                out.print( " " );
            for ( int i = 0; i < count; i++ )
                out.print( "\b" );
            out.flush();
        }
        catch ( Exception e ) {
            e.printStackTrace();
        }
    }

    /** Makes a nifty ascii spinner */
    private void startSpinner() {
        canSpin = true;
        spinner =
            new Thread() {
                public void run() {
                    try {
                        if ( doSpin ) {

                            print( working );
                            String time = "";
                            if ( stopwatch != null ) {
                                stopwatch.reset();
                                time = stopwatch.format( stopwatch.elapsed() );
                            }
                            while ( canSpin ) {
                                if ( stopwatch != null ) {
                                    time = stopwatch.format( stopwatch.elapsed() ) + " ";
                                    print( time );
                                }
                                print( "|" );
                                Thread.currentThread().sleep( 200 );
                                erase( 1 );
                                print( "/" );
                                Thread.currentThread().sleep( 200 );
                                erase( 1 );
                                print( "-" );
                                Thread.currentThread().sleep( 200 );
                                erase( 1 );
                                print( "\\" );
                                Thread.currentThread().sleep( 200 );
                                erase( 1 );
                                erase( time.length() );
                            }
                        }
                    }
                    catch ( Exception e ) {
                        //e.printStackTrace();
                    }
                    erase( working.length() + 1 );
                }
            };
        spinner.setPriority( Thread.NORM_PRIORITY - 1 );
        spinner.start();
    }

    /** Stops the spinner */
    private void stopSpinner() {
        canSpin = false;
        if ( spinner != null ) {
            spinner.interrupt();
        }
    }

    /**
     * Sets the echo attribute of the DbConsole object
     *
     * @param b  The new echo value
     */
    public void setEcho( boolean b ) {
        echo = b;
    }

    /**
     * Gets the echo attribute of the DbConsole object
     *
     * @return   The echo value
     */
    public boolean getEcho() {
        return echo;
    }


    /**
     * Sets the output stream of the console to the given Writer. This is useful
     * for writing the output to a file, for example. See the FileOutputCommand
     * for a working example.
     *
     * @param new_out  The new out value
     */
    public void setOut( PrintStream new_out ) {
        if ( new_out == null )
            throw new IllegalArgumentException( "Writer may not be null." );
        out = new_out;
    }

    /**
     * Gets the current output PrintStream. Note that System.out is the default,
     * it is generally not a good idea to close System.out.
     *
     * @return   The out value
     */
    public PrintStream getOut() {
        return out;
    }

    /** Resets the default output Writer to System.out. */
    public void resetOut() {
        out = _system_out;
    }

    /**
     * Sets the in attribute of the DbConsole class
     *
     * @param is  The new in value
     */
    public void setIn( InputStream is ) {
        if ( is == null )
            throw new IllegalArgumentException( "InputStream may not be null." );
        if ( is != _system_in && _system_in.markSupported() )
            _system_in.mark( 256 );
        in = is;
        startCommandLoop();
    }

    /** Description of the Method */
    public void startCommandLoop() {
        stopCommandLoop();
        canLoop = true;
        createCommandLoop();
        commandLoop.start();
    }

    /** Description of the Method */
    public void stopCommandLoop() {
        try {
            canLoop = false;
            if ( commandLoop != null ) {
                commandLoop.interrupt();
                commandLoop.join();
            }
        }
        catch ( InterruptedException ignored ) {}
    }

    /**
     * Gets the in attribute of the DbConsole class
     *
     * @return   The in value
     */
    public InputStream getIn() {
        return in;
    }

    /** Resets the default input reader to System.in. */
    public void resetIn() {
        try {
            if ( in != _system_in )
                in.close();
        }
        catch ( Exception e ) {
            e.printStackTrace();
        }
        in = _system_in;
        try {
            if ( in.markSupported() )
                in.reset();
        }
    catch ( Exception e ) {}
        startCommandLoop();
    }

    /**
     * Executes a command. A command may have data following it, for example
     * "select * from mytable". In this case, 'select' is the command. The full
     * command will be passed to the command processor, continuing the example,
     * all of "select * from mytable" will be passed.
     *
     * @param command  the command to execute
     */
    public void executeCommand( String command ) {
        if ( command == null || command.length() == 0 )
            return ;
        try {
            // can have multiple commands separated by ;
            String[] cmds = command.split( "[;]" );
            for ( int i = 0; i < cmds.length; i++ ) {
                // get the command name
                cmds[ i ] = cmds[ i ].trim();
                String[] args = cmds[ i ].split( "[ ]" );
                if ( args.length == 0 )
                    continue;
                args[ 0 ] = args[ 0 ].trim();
                if ( args[ 0 ].length() == 0 )
                    continue;

                // get the actual Command to execute
                String classname = command_lu.getProperty( args[ 0 ].toLowerCase() );
                Command cmd = ( Command ) commands.get( classname );
                if ( cmd == null ) {
                    try {
                        cmd = ( Command ) Class.forName( classname ).newInstance();
                        commands.put( classname, cmd );
                    }
                    catch ( Exception e ) {
                        //System.out.println( args[ 0 ] + " " + classname );
                        //e.printStackTrace();
                        println( msg_prompt + new ColoredString( Bundle.BUNDLE.getString( "unknownCommand" ), AnsiColor.WARN_COLOR ) );
                        continue;
                    }
                    if ( cmd == null ) {
                        println( msg_prompt + new ColoredString( Bundle.BUNDLE.getString( "unknownCommand" ), AnsiColor.WARN_COLOR ) );
                        continue;
                    }
                }

                // execute the command
                Object output = null;
                if ( cmd.showSpinner() ) {
                    startSpinner();
                    stopwatch = new StopWatch();
                }
                if ( args.length == 1 )
                    output = cmd.execute();
                else
                    output = cmd.execute( cmds[ i ] );
                if ( cmd.showSpinner() ) {
                    stopSpinner();
                    try {
                        spinner.join();
                    }
                    catch ( Exception e ) {
                        // ignored, was just waiting for the spinner thread to quit
                    }
                }

                // show any output
                StringBuffer sb = new StringBuffer();
                sb.append( LS );
                if ( output != null )
                    sb.append( output.toString() );
                if ( cmd.showSpinner() ) {
                    stopwatch.stop();
                    sb.append( LS );
                    sb.append( stopwatch.format( stopwatch.total() ) );
                    sb.append( LS );
                }
                println( sb.toString() );
            }
        }
        catch ( Exception e ) {
            e.printStackTrace();
            println( msg_prompt + new ColoredString( Bundle.BUNDLE.getString( "error" ), AnsiColor.ERROR_COLOR ) + " " + e.getMessage() );
        }
    }

    class CommandThread extends Thread {

        private String command;

        /** Constructor for CommandThread */
        public CommandThread( String cmd ) {
            setPriority( Thread.NORM_PRIORITY - 1 );
            command = cmd;
        }

        public void setCommand( String cmd ) {
            command = cmd;
        }

        public String getCommand() {
            return command;
        }

        public void run() {
            DbConsoleDAO.CAN_RUN = true;
            executeCommand( command );
            DbConsole.this.lastCommand = command;
            if ( DbConsoleDAO.CAN_RUN == false ) {
                println( msg_prompt + Bundle.BUNDLE.getString( "interrupted" ) );
            }
            println();
        }

        public void interrupt() {
            stopSpinner();
            DbConsoleDAO.CAN_RUN = false;
            super.interrupt();
            println();
        }
    }

    /**
     * The main program for the DbConsole class
     *
     * @param args  The command line arguments, currently none are recognized.
     */
    public static void main( String[] args ) {
        new DbConsole();
    }
}
