/*
Copyright (c) Dale Anson, 2004
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote products
derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package ise.dbconsole;

import java.io.*;
import java.sql.*;
import java.util.*;
import javax.sql.DataSource;
import ise.dbconsole.to.*;
import ise.dbconsole.exception.*;

import ise.library.*;
import ise.library.ascii.*;

import org.apache.commons.dbcp.ConnectionFactory;
import org.apache.commons.dbcp.DriverConnectionFactory;
import org.apache.commons.dbcp.PoolableConnectionFactory;
import org.apache.commons.dbcp.PoolingDataSource;
import org.apache.commons.pool.ObjectPool;
import org.apache.commons.pool.impl.GenericObjectPool;

/**
 * Simple, reusable DAO. All transfer objects are Strings (simple, eh?)
 *
 * @author    Dale Anson, Sep 2004
 * @version   $Revision: 18 $
 */
public class DbConsoleDAO implements Constants {

    /** the data source */
    private DataSource dataSource = null;

    private DbInfo dbInfo = null;

    /** the database driver */
    private String dbDriver;

    /** the url to connect to the database */
    private String connectionUrl;

    /** a connection to use when auto commit is off */
    private Connection autoCommitConnection = null;

    /** whether to show raw binary data */
    private boolean rawMode = false;

    /** landscape mode on or off */
    private boolean landscapeMode = false;

    /** auto commit on or off */
    private boolean autoCommitMode = true;

    private String defaultCatalog = null;
    private String defaultSchema = null;
    private TableArray table = null;

    public static volatile boolean CAN_RUN = true;


    /**
     * Constructor for DbConsoleDAO
     *
     * @param connectionUrl  the database connection url
     * @param dbDriver       the database driver class name
     * @param props          properties to pass to the driver
     * @exception Exception  Description of Exception
     * @deprecated
     */
    public DbConsoleDAO( String dbDriver, String connectionUrl, Properties props ) throws Exception {   // NOPMD
        throw new UnsupportedOperationException( "deprecated, use DbConsoleDAO(DbInfo)" );
    }

    /**
     * Constructor for DbConsoleDAO
     *
     * @param dbinfo
     * @param props
     * @exception Exception  Description of Exception
     * @deprecated
     */
    public DbConsoleDAO( DbInfo dbinfo, Properties props ) throws Exception {   // NOPMD
        throw new UnsupportedOperationException( "deprecated, use DbConsoleDAO(DbInfo)" );
    }

    /**
     * Constructor for DbConsoleDAO
     *
     * @param dbinfo
     * @exception Exception  Description of Exception
     */
    public DbConsoleDAO( DbInfo dbinfo ) throws Exception {
        this.dbInfo = dbinfo;
        this.dbDriver = dbinfo.getDriver();
        this.connectionUrl = dbinfo.getUrl();
        Properties user = dbinfo.getUserProperties();

        // load the underlying JDBC driver
        Class driver_class = Class.forName( dbDriver, true, Thread.currentThread().getContextClassLoader() );
        Driver driver = ( Driver ) driver_class.newInstance();
        DriverManager.registerDriver( driver );

        // set up the PoolingDataSource
        ObjectPool connectionPool = new GenericObjectPool( null );
        ConnectionFactory connectionFactory = new DriverConnectionFactory( driver, connectionUrl, user );
        PoolableConnectionFactory poolableConnectionFactory = new PoolableConnectionFactory( connectionFactory, connectionPool, null, null, false, true );
        dataSource = new PoolingDataSource( connectionPool );
    }

    /**
     * Gets the currentDbInfo attribute of the DbConsoleDAO object
     *
     * @return   The currentDbInfo value
     */
    public DbInfo getCurrentDbInfo() {
        return dbInfo;
    }


    /**
     * Gets the current database connection url.
     *
     * @return   The url value
     */
    public String getUrl() {
        return connectionUrl;
    }

    /**
     * Gets the classname of the current database driver.
     *
     * @return   The driverClassName value
     */
    public String getDriverClassName() {
        return dbDriver;
    }

    /**
     * Gets a little bit of information about the database itself.
     *
     * @return   The dbInfo value
     */
    public String getDbInfo() {
        Connection conn = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
        StringBuffer sb = new StringBuffer();

        try {
            conn = getConnection();
            DatabaseMetaData dmd = conn.getMetaData();
            try {
                sb.append( "Url: " ).append( dmd.getURL() ).append( LS );
                sb.append( "Driver: " ).append( dbDriver ).append( LS );
                sb.append( "Username: " ).append( dmd.getUserName() ).append( LS );
            }
            catch ( Exception e ) {}
            try {
                sb.append( "Database Product Name: " ).append( dmd.getDatabaseProductName() ).append( LS );
            }
            catch ( Exception e ) {}
            try {
                sb.append( "Database Product Version: " ).append( dmd.getDatabaseProductVersion() ).append( LS );
            }
            catch ( Exception e ) {}
            return sb.toString();
        }
        catch ( Exception e ) {
            return Bundle.BUNDLE.getString( "messagePrompt" ) + " " + new ColoredString( Bundle.BUNDLE.getString( "error" ), AnsiColor.ERROR_COLOR ) + " " + e.getMessage();
        }
        finally {
            closeResources( conn, stmt, rs );
        }
    }

    /**
     * Gets a pile of information about the database itself.
     *
     * @return   The dbInfo value
     */
    public String getDbInformation() {
        Connection conn = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
        StringBuffer sb = new StringBuffer();

        TableArray results = new TableArray();

        try {
            conn = getConnection();
            DatabaseMetaData dmd = conn.getMetaData();
            int y = 0;
            try {
                sb.append( "Url: " ).append( dmd.getURL() ).append( LS );
                sb.append( "Driver: " ).append( dbDriver ).append( LS );
                sb.append( "Username: " ).append( dmd.getUserName() ).append( LS );
            }
            catch ( Exception e ) {}
            try {
                sb.append( "Database Product Name: " ).append( dmd.getDatabaseProductName() ).append( LS );
            }
            catch ( Exception e ) {}
            try {
                sb.append( "Database Product Version: " ).append( dmd.getDatabaseProductVersion() ).append( LS );
            }
            catch ( Exception e ) {}

            // header

            results.put( 1, y, "Property" );
            results.put( 2, y, "Value" );

            // info
            ++y;
            try {
                results.put( 1, y, "Procedures callable" );
                results.put( 2, y, String.valueOf( dmd.allProceduresAreCallable() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Tables selectable" );
                results.put( 2, y, String.valueOf( dmd.allTablesAreSelectable() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Data def causes transaction commit" );
                results.put( 2, y, String.valueOf( dmd.dataDefinitionCausesTransactionCommit() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Data def ignored in transactions" );
                results.put( 2, y, String.valueOf( dmd.dataDefinitionIgnoredInTransactions() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Deletes detected, forward only" );
                results.put( 2, y, String.valueOf( dmd.deletesAreDetected( ResultSet.TYPE_FORWARD_ONLY ) ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Deletes detected, scroll insensitive" );
                results.put( 2, y, String.valueOf( dmd.deletesAreDetected( ResultSet.TYPE_SCROLL_INSENSITIVE ) ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Deletes detected, scroll sensitive" );
                results.put( 2, y, String.valueOf( dmd.deletesAreDetected( ResultSet.TYPE_SCROLL_SENSITIVE ) ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Max row size includes blobs" );
                results.put( 2, y, String.valueOf( dmd.doesMaxRowSizeIncludeBlobs() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Catalog Separator" );
                results.put( 2, y, String.valueOf( dmd.getCatalogSeparator() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Catalog Term" );
                results.put( 2, y, String.valueOf( dmd.getCatalogTerm() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Database Major Version" );
                results.put( 2, y, String.valueOf( dmd.getDatabaseMajorVersion() ) );
                ++y;
            }
            catch ( Throwable e ) {     // NOPMD
                // throwable because Oracle's driver throws an AbstractMethod Error
            }
            try {
                results.put( 1, y, "Database Minor Version" );
                results.put( 2, y, String.valueOf( dmd.getDatabaseMinorVersion() ) );
                ++y;
            }
            catch ( Throwable e ) {     // NOPMD
                // throwable because Oracle's driver throws an AbstractMethod Error
            }
            try {
                results.put( 1, y, "Default Transaction Isolation" );
                results.put( 2, y, String.valueOf( dmd.getDefaultTransactionIsolation() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Driver Major Version" );
                results.put( 2, y, String.valueOf( dmd.getDriverMajorVersion() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Driver Minor Version" );
                results.put( 2, y, String.valueOf( dmd.getDriverMinorVersion() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Driver Name" );
                results.put( 2, y, String.valueOf( dmd.getDriverName() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Driver Version" );
                results.put( 2, y, String.valueOf( dmd.getDriverVersion() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Extra Name Characters" );
                results.put( 2, y, String.valueOf( dmd.getExtraNameCharacters() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Identifier Quote String" );
                results.put( 2, y, String.valueOf( dmd.getIdentifierQuoteString() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "JDBC Major Version" );
                results.put( 2, y, String.valueOf( dmd.getJDBCMajorVersion() ) );
                ++y;
            }
            catch ( Throwable e ) {     // NOPMD
                // throwable because Oracle's driver throws an AbstractMethod Error
            }
            try {
                results.put( 1, y, "JDBC Minor Version" );
                results.put( 2, y, String.valueOf( dmd.getJDBCMinorVersion() ) );
                ++y;
            }
            catch ( Throwable e ) {     // NOPMD
                // throwable because Oracle's driver throws an AbstractMethod Error
            }
            try {   
                results.put( 1, y, "Max Binary Literal Length" );
                results.put( 2, y, String.valueOf( dmd.getMaxBinaryLiteralLength() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Max Catalog Name Length" );
                results.put( 2, y, String.valueOf( dmd.getMaxCatalogNameLength() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Max Char Literal Length" );
                results.put( 2, y, String.valueOf( dmd.getMaxCharLiteralLength() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Max Column Name Length" );
                results.put( 2, y, String.valueOf( dmd.getMaxColumnNameLength() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Max Columns In Group By" );
                results.put( 2, y, String.valueOf( dmd.getMaxColumnsInGroupBy() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Max Columns In Index" );
                results.put( 2, y, String.valueOf( dmd.getMaxColumnsInIndex() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Max Columns In Order By" );
                results.put( 2, y, String.valueOf( dmd.getMaxColumnsInOrderBy() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Max Columns In Select" );
                results.put( 2, y, String.valueOf( dmd.getMaxColumnsInSelect() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Max Columns In Table" );
                results.put( 2, y, String.valueOf( dmd.getMaxColumnsInTable() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Max Connections" );
                results.put( 2, y, String.valueOf( dmd.getMaxConnections() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Max Cursor Name Length" );
                results.put( 2, y, String.valueOf( dmd.getMaxCursorNameLength() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Max Index Length" );
                results.put( 2, y, String.valueOf( dmd.getMaxIndexLength() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Max Procedure Name Length" );
                results.put( 2, y, String.valueOf( dmd.getMaxProcedureNameLength() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Max Row Size" );
                results.put( 2, y, String.valueOf( dmd.getMaxRowSize() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Max Schema Name Length" );
                results.put( 2, y, String.valueOf( dmd.getMaxSchemaNameLength() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Max Statement Length" );
                results.put( 2, y, String.valueOf( dmd.getMaxStatementLength() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Max Statements" );
                results.put( 2, y, String.valueOf( dmd.getMaxStatements() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Max Table Name Length" );
                results.put( 2, y, String.valueOf( dmd.getMaxTableNameLength() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Max Tables In Select" );
                results.put( 2, y, String.valueOf( dmd.getMaxTablesInSelect() ) );
                ++y;
            }
            catch ( Exception e ) {}
            try {
                results.put( 1, y, "Max User Name Length" );
                results.put( 2, y, String.valueOf( dmd.getMaxUserNameLength() ) );
                ++y;
            }
            catch ( Exception e ) {}

            sb.append( LS ).append( tableArrayToFormattedString( results ) );
            try {
                sb.append( LS );
                sb.append( "Numeric Functions: " ).append( dmd.getNumericFunctions() ).append( LS );
                sb.append( LS );
                sb.append( "SQL Key Words: " ).append( dmd.getSQLKeywords() ).append( LS );
                sb.append( LS );
                sb.append( "String Functions: " ).append( dmd.getStringFunctions() ).append( LS );
                sb.append( LS );
                sb.append( "System Functions: " ).append( dmd.getSystemFunctions() ).append( LS );
                sb.append( LS );
                sb.append( "Time and Date Functions: " ).append( dmd.getTimeDateFunctions() ).append( LS );

            }
            catch ( Exception e ) {}

            return sb.toString();
        }
        catch ( Exception e ) {
            //e.printStackTrace();
            return Bundle.BUNDLE.getString( "messagePrompt" ) + " " + new ColoredString( Bundle.BUNDLE.getString( "error" ), AnsiColor.ERROR_COLOR ) + " " + e.getMessage();
        }
        finally {
            closeResources( conn, stmt, rs );
        }
    }

    /**
     * Gets a list of schemas in the database.
     *
     * @return   The schemas
     */
    public String getSchemas() {
        Connection conn = null;
        ResultSet rs = null;

        try {
            conn = getConnection();
            DatabaseMetaData dmd = conn.getMetaData();
            rs = dmd.getSchemas();
            return tableArrayToFormattedString( resultSetToTableArray( rs ) );
        }
        catch ( Exception e ) {
            return Bundle.BUNDLE.getString( "messagePrompt" ) + " " + new ColoredString( Bundle.BUNDLE.getString( "error" ), AnsiColor.ERROR_COLOR ) + " " + e.getMessage();
        }
        finally {
            closeResources( conn, rs );
        }
    }

    /**
     * Gets a list of tables in the database, can be filtered by catalog,
     * schema, and table pattern.
     *
     * @param catalog_name   the name of the catalog, use null for all or "" for
     *      tables without a catalog.
     * @param table_pattern  exact name or pattern, use % as pattern character.
     * @param schema_name
     * @return               The tables value
     */
    public String getTables( String catalog_name, String schema_name, String table_pattern ) {
        Connection conn = null;
        ResultSet rs = null;

        if ( catalog_name == null )
            catalog_name = defaultCatalog;
        if ( schema_name == null )
            schema_name = defaultSchema;
        if ( table_pattern == null )
            table_pattern = "%";
        //System.out.println(String.valueOf(catalog_name) + "." + String.valueOf(schema_name) + "." + table_pattern);
        try {
            conn = getConnection();
            DatabaseMetaData dmd = conn.getMetaData();
            if ( dmd == null )
                throw new RuntimeException( Bundle.BUNDLE.getString( "cantGetDatabaseMetaData" ) );
            rs = dmd.getTables( catalog_name, schema_name, table_pattern, null );
            TableArray ta = resultSetToTableArray( rs );
            if ( ta.getRowCount() == 1 ) {
                // try again using uppercase, all we got was the column headers
                catalog_name = catalog_name == null ? null : catalog_name.toUpperCase();
                schema_name = schema_name == null ? null : schema_name.toUpperCase();
                table_pattern = table_pattern.toUpperCase();
                rs = dmd.getTables( catalog_name, schema_name, table_pattern, null );
                ta = resultSetToTableArray( rs );
            }
            return tableArrayToFormattedString( ta );
        }
        catch ( Exception e ) {
            e.printStackTrace();
            return Bundle.BUNDLE.getString( "messagePrompt" ) + " " + new ColoredString( Bundle.BUNDLE.getString( "error" ), AnsiColor.ERROR_COLOR ) + " " + e.getMessage();
        }
        finally {
            closeResources( conn, rs );
        }
    }

    /**
     * Gets the procedures attribute of the DbConsoleDAO object
     *
     * @param catalog_name
     * @param schema_pattern
     * @param proc_pattern
     * @return                The procedures value
     */
    public String getProcedures( String catalog_name, String schema_pattern, String proc_pattern ) {
        Connection conn = null;
        ResultSet rs = null;

        if ( catalog_name == null )
            catalog_name = defaultCatalog;
        if ( schema_pattern == null && defaultSchema == null )
            schema_pattern = "%";
        if ( schema_pattern == null && defaultSchema != null )
            schema_pattern = defaultSchema;
        if ( proc_pattern == null )
            proc_pattern = "%";
        try {
            conn = getConnection();
            DatabaseMetaData dmd = conn.getMetaData();
            rs = dmd.getProcedures( catalog_name, schema_pattern, proc_pattern );
            return tableArrayToFormattedString( resultSetToTableArray( rs ) );
        }
        catch ( Exception e ) {
            return Bundle.BUNDLE.getString( "messagePrompt" ) + " " + new ColoredString( Bundle.BUNDLE.getString( "error" ), AnsiColor.ERROR_COLOR ) + " " + e.getMessage();
        }
        finally {
            closeResources( conn, rs );
        }
    }

    /**
     * Gets the columns of the tables matching the schema name and table
     * pattern. This returns a very verbose and detailed listing of column
     * attributes.
     *
     * @param schema_name
     * @param table_pattern
     * @param column_pattern
     * @return                The columns value
     */
    public String getColumns( String schema_name, String table_pattern, String column_pattern ) {
        return getColumns( null, schema_name, table_pattern, column_pattern );
    }

    public String getColumns( String catalog_name, String schema_name, String table_pattern, String column_pattern ) {
        Connection conn = null;
        ResultSet rs = null;

        if ( catalog_name == null )
            catalog_name = defaultCatalog;
        if ( schema_name == null )
            schema_name = defaultSchema;
        if ( table_pattern == null )
            table_pattern = "%";
        if ( column_pattern == null )
            column_pattern = "%";
        try {
            conn = getConnection();
            DatabaseMetaData dmd = conn.getMetaData();
            rs = dmd.getColumns( catalog_name, schema_name, table_pattern, column_pattern );
            TableArray ta = resultSetToTableArray( rs );
            if ( ta.getRowCount() == 1 ) {
                // try again using uppercase, all we got was the column headers
                catalog_name = catalog_name == null ? null : catalog_name.toUpperCase();
                schema_name = schema_name == null ? null : schema_name.toUpperCase();
                table_pattern = table_pattern.toUpperCase();
                column_pattern = column_pattern.toUpperCase();
                rs = dmd.getColumns( catalog_name, schema_name, table_pattern, column_pattern );
                ta = resultSetToTableArray( rs );
            }
            return tableArrayToFormattedString( ta );
        }
        catch ( Exception e ) {
            return Bundle.BUNDLE.getString( "messagePrompt" ) + " " + new ColoredString( Bundle.BUNDLE.getString( "error" ), AnsiColor.ERROR_COLOR ) + " " + e.getMessage();
        }
        finally {
            closeResources( conn, rs );
        }
    }

    /**
     * Lists the column attributes for the given schema and table pattern. This
     * is not as verbose as <code>getColumns</code> and is more useful for a
     * quick look at the columns in a table. Only 5 attributes per column are
     * shown: column name, data type, type name, decimal digits, and is
     * nullable.
     *
     * @param schema_name
     * @param table_pattern
     * @return               Description of the Returned Value
     */
    public String listColumns( String schema_name, String table_pattern ) {
        return listColumns( schema_name, table_pattern, false );
    }

    /**
     * Lists the column attributes for the given schema and table pattern. This
     * is not as verbose as <code>getColumns</code> and is more useful for a
     * quick look at the columns in a table. Only 5 attributes per column are
     * shown: column name, data type, type name, decimal digits, and is
     * nullable.
     *
     * @param schema_name
     * @param table_pattern
     * @param brief if true, only show column name and compact type
     * @return               Description of the Returned Value
     */
    public String listColumns( String schema_name, String table_pattern, boolean brief ) {
        Connection conn = null;
        ResultSet rs = null;

        if ( schema_name == null )
            schema_name = defaultSchema;
        if ( table_pattern == null )
            table_pattern = "%";
        try {
            conn = getConnection();
            DatabaseMetaData dmd = conn.getMetaData();
            if ( schema_name != null )
                schema_name = schema_name.toUpperCase();   /// check that this and the next line work everywhere
            table_pattern = table_pattern.toUpperCase();
            rs = dmd.getColumns( null, schema_name, table_pattern, "%" );
            int y = 0;
            TableArray table = new TableArray();

            StringBuffer sb = new StringBuffer();

            // set up columns and headers
            ResultSetMetaData md = rs.getMetaData();
            String col1 = md.getColumnName( 4 ).toUpperCase();   // column name
            String col2 = md.getColumnName( 6 ).toUpperCase();   // data type name
            String col3 = md.getColumnName( 7 ).toUpperCase();   // column width
            String col4 = md.getColumnName( 13 ).toUpperCase();   // column default
            String col5 = md.getColumnName( 18 ).toUpperCase();   // is nullable
            if ( brief ) {
                table.put( 1, y, "NAME" );
                table.put( 2, y, "TYPE" );
            }
            else {
                table.put( 1, y, col1 );
                table.put( 2, y, col2 );
                table.put( 3, y, col3 );
                table.put( 4, y, col4 );
                table.put( 5, y, col5 );
            }
            ++y;

            String new_schema_name = null;
            String new_table_name = null;
            schema_name = null;
            String table_name = null;
            Object value;
            String svalue;
            Object size;
            while ( rs.next() ) {
                value = rs.getObject( 2 );
                new_schema_name = value == null ? "" : value.toString();
                if ( schema_name == null )
                    schema_name = new_schema_name;
                if ( !new_schema_name.equals( schema_name ) ) {
                    sb.append( schema_name ).append( schema_name.length() > 0 ? "." : "" ).append( table_name ).append( LS );
                    sb.append( tableArrayToFormattedString( table ) ).append( LS ).append( LS );
                    table = new TableArray();
                    y = 0;
                    if ( brief ) {
                        table.put( 1, y, col1 );
                        table.put( 2, y, col2 + "(" + col3 + ")" );
                    }
                    else {
                        table.put( 1, y, col1 );
                        table.put( 2, y, col2 );
                        table.put( 3, y, col3 );
                        table.put( 4, y, col4 );
                        table.put( 5, y, col5 );
                    }
                    ++y;
                    schema_name = new_schema_name;
                    value = rs.getObject( 3 );
                    table_name = value == null ? "" : value.toString();
                }
                value = rs.getObject( 3 );
                new_table_name = value == null ? "" : value.toString();
                if ( table_name == null )
                    table_name = new_table_name;
                if ( !new_table_name.equals( table_name ) ) {
                    sb.append( schema_name ).append( schema_name.length() > 0 ? "." : "" ).append( table_name ).append( LS );
                    sb.append( tableArrayToFormattedString( table ) ).append( LS ).append( LS );
                    table = new TableArray();
                    y = 0;
                    if ( brief ) {
                        table.put( 1, y, "NAME" );
                        table.put( 2, y, "TYPE" );
                    }
                    else {
                        table.put( 1, y, col1 );
                        table.put( 2, y, col2 );
                        table.put( 3, y, col3 );
                        table.put( 4, y, col4 );
                        table.put( 5, y, col5 );
                    }
                    ++y;
                    table_name = new_table_name;
                }
                if ( brief ) {
                    value = rs.getObject( 4 );
                    table.put( 1, y, value == null ? "" : value.toString().trim() );
                    value = rs.getObject( 6 );
                    svalue = value == null ? "" : value.toString().trim();
                    size = rs.getObject( 7 );
                    svalue += "(" + ( size == null ? "?" : size.toString().trim() ) + ")";
                    table.put( 2, y, svalue );
                }
                else {
                    value = rs.getObject( 4 );
                    table.put( 1, y, value == null ? "" : value.toString().trim() );
                    value = rs.getObject( 6 );
                    table.put( 2, y, value == null ? "" : value.toString().trim() );
                    value = rs.getObject( 7 );
                    table.put( 3, y, value == null ? "" : value.toString().trim() );
                    value = rs.getObject( 13 );
                    table.put( 4, y, value == null ? "" : value.toString().trim() );
                    value = rs.getObject( 18 );
                    table.put( 5, y, value == null ? "" : value.toString().trim() );
                }
                ++y;
            }
            if ( schema_name == null )
                schema_name = ""; //throw new Exception("Schema name is missing.");
            sb.append( schema_name ).append( schema_name.length() > 0 ? "." : "" ).append( table_name ).append( LS );
            return sb.append( tableArrayToFormattedString( table ) ).toString();
        }
        catch ( Exception e ) {
            e.printStackTrace();
            return Bundle.BUNDLE.getString( "messagePrompt" ) + " " + new ColoredString( Bundle.BUNDLE.getString( "error" ), AnsiColor.ERROR_COLOR ) + " " + e.getMessage();
        }
        finally {
            closeResources( conn, rs );
        }
    }

    /**
     * Gets the tablePrivileges attribute of the DbConsoleDAO object
     *
     * @param schema_name
     * @param table_pattern
     * @return               The tablePrivileges value
     */
    public String getTablePrivileges( String schema_name, String table_pattern ) {
        Connection conn = null;
        ResultSet rs = null;

        if ( schema_name == null )
            schema_name = defaultSchema;
        if ( table_pattern == null )
            table_pattern = "%";
        try {
            conn = getConnection();
            DatabaseMetaData dmd = conn.getMetaData();
            rs = dmd.getTablePrivileges( null, schema_name, table_pattern );
            TableArray ta = resultSetToTableArray( rs );
            if ( ta.getRowCount() == 1 ) {
                // try again with uppercase, all we got was the column headers
                schema_name = schema_name == null ? null : schema_name.toUpperCase();
                table_pattern = table_pattern.toUpperCase();
                rs = dmd.getTablePrivileges( null, schema_name, table_pattern );
                ta = resultSetToTableArray( rs );
            }
            return tableArrayToFormattedString( ta );
        }
        catch ( Exception e ) {
            return Bundle.BUNDLE.getString( "messagePrompt" ) + " " + new ColoredString( Bundle.BUNDLE.getString( "error" ), AnsiColor.ERROR_COLOR ) + " " + e.getMessage();
        }
        finally {
            closeResources( conn, rs );
        }
    }

    /**
     * Gets the connection attribute of the DbConsole object
     *
     * @return                  The connection value
     * @exception SQLException  Description of Exception
     */
    public Connection getConnection() throws SQLException {
        if ( autoCommitMode ) {
            return dataSource.getConnection();
        }
        else {
            if ( autoCommitConnection == null ) {
                autoCommitConnection = dataSource.getConnection();
                autoCommitConnection.setAutoCommit( false );
            }
            return autoCommitConnection;
        }
    }

    public String commit() {
        if ( autoCommitConnection != null ) {
            try {
                autoCommitConnection.commit();
                closeConnection( autoCommitConnection );
                autoCommitConnection = null;
                return Bundle.BUNDLE.getString( "messagePrompt" ) + " " + Bundle.BUNDLE.getString( "committed" );
            }
            catch ( Exception e ) {
                return Bundle.BUNDLE.getString( "messagePrompt" ) + " " + new ColoredString( Bundle.BUNDLE.getString( "error" ), AnsiColor.ERROR_COLOR ) + " " + e.getMessage();
            }
        }
        return "";
    }

    /**
     * Does an sql select and returns the results.
     *
     * @param sql
     * @return     Description of the Returned Value
     */
    public String doSelect( String sql ) {
        Connection conn = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;

        try {
            conn = getConnection();
            stmt = conn.prepareStatement( sql );
            rs = stmt.executeQuery();
            System.out.println( "\nquery complete, formatting output..." );
            if ( !CAN_RUN )
                return Bundle.BUNDLE.getString( "messagePrompt" ) + " " + Bundle.BUNDLE.getString( "interrupted" );
            if ( rawMode )
                return resultSetToRaw( rs );
            return tableArrayToFormattedString( resultSetToTableArray( rs ) );
        }
        catch ( Exception e ) {
            return Bundle.BUNDLE.getString( "messagePrompt" ) + " " + new ColoredString( Bundle.BUNDLE.getString( "error" ), AnsiColor.ERROR_COLOR ) + " " + e.getMessage();
        }
        finally {
            closeResources( conn, stmt, rs );
        }
    }

    /**
     * Shows the detail of a single row from the previous results, if any.  The
     * returned string will be in a two-column format, with the left column containing
     * the column names and the right column containing the row data.  This makes it
     * easier to examine the data for a single row.
     * @param row the index of the row from the previous results
     */
    public String getRow( int row ) {
        if ( table == null )
            return null;
        if ( row < 1 || row > table.getRowCount() )
            throw new RowIndexOutOfBoundsException();

        TableArray ta = new TableArray();

        List col_names = new ArrayList();
        col_names.add( "COL NAME" );
        List names = table.getRow( 0 );

        // maybe remove the ROW column
        if ( YES.equals( System.getProperty( SHOW_ROW_COLUMN ) ) )
            names.remove( 0 );
        col_names.addAll( names );
        ta.putColumn( 1, col_names );

        List row_data = new ArrayList();
        row_data.add( "DATA" );
        List data = table.getRow( row );

        // maybe remove the ROW column
        if ( YES.equals( System.getProperty( SHOW_ROW_COLUMN ) ) )
            data.remove( 0 );

        row_data.addAll( data );
        ta.putColumn( 2, row_data );

        return tableArrayToFormattedString( ta );

    }

    /**
     * Rotates the result set into landscape mode, that is, rotates the output
     * table 90 degrees counter-clockwise so the column names are in the left column,
     * and data rows become data columns.
     */
    private TableArray convertToLandscape() {
        if ( table == null )
            return null;

        TableArray ta = new TableArray();

        List col_names = new ArrayList();
        col_names.add( "COL NAME" );
        List names = table.getRow( 0 );

        // maybe remove the ROW column
        if ( YES.equals( System.getProperty( SHOW_ROW_COLUMN ) ) )
            names.remove( 0 );
        col_names.addAll( names );
        ta.putColumn( 1, col_names );

        for ( int row = 1; row < table.getRowCount(); row++ ) {
            List row_data = new ArrayList();
            row_data.add( "ROW " + row );
            List data = table.getRow( row );

            // maybe remove the ROW column
            if ( YES.equals( System.getProperty( SHOW_ROW_COLUMN ) ) )
                data.remove( 0 );

            row_data.addAll( data );
            ta.putColumn( row + 1, row_data );
        }
        return ta;
    }

    /**
     * Turns auto commit on or off.  If off, all sql commands will reuse the
     * same connection.  A 'commit' command must be executed at some point for
     * the data to be committed to the db.
     *
     * @param auto  The new autoCommitMode value
     */
    public void setAutoCommitMode( boolean auto ) {
        autoCommitMode = auto;
    }

    /**
     * Gets the autoCommitMode attribute of the DbConsoleDAO object
     *
     * @return   The autoCommitMode value
     */
    public boolean getAutoCommitMode() {
        return autoCommitMode;
    }

    /**
     * Sets whether a 'select' should get the raw data from a binary object. Use
     * 'true' to set raw mode, use 'false' to use regular mode. When in raw
     * mode, the output will not be tabulated, so this is useful for saving
     * binary data directly to a file.
     *
     * @param raw  The new rawMode value
     */
    public void setRawMode( boolean raw ) {
        rawMode = raw;
    }

    /**
     * Gets the rawMode attribute of the DbConsoleDAO object
     *
     * @return   The rawMode value
     */
    public boolean getRawMode() {
        return rawMode;
    }

    public void setLandscapeMode( boolean landscape ) {
        landscapeMode = landscape;
    }

    public boolean getLandscapeMode() {
        return landscapeMode;
    }

    public void setDefaultSchema( String schema ) {
        defaultSchema = schema;
    }

    public String getDefaultSchema() {
        return defaultSchema;
    }

    public void setDefaultCatalog( String catalog ) {
        defaultCatalog = catalog;
    }

    public String getDefaultCatalog() {
        return defaultCatalog;
    }

    /**
     * Executes some data manipulation call, e.g. update, insert, or delete.
     * Should not use this method for 'select' as this method only prints out
     * the number of rows altered. For 'select', use the 'doSelect' method.
     *
     * @param sql
     * @return     Description of the Returned Value
     */
    public String doDML( String sql ) {
        Connection conn = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;

        try {
            conn = getConnection();
            stmt = conn.prepareStatement( sql );
            boolean has_result_set = stmt.execute();
            if ( has_result_set ) {
                rs = stmt.getResultSet();
                TableArray ta = resultSetToTableArray( rs );
                return tableArrayToFormattedString( ta );
            }
            else {
                int r = stmt.getUpdateCount();
                return Bundle.BUNDLE.getString( "messagePrompt" ) + " Modified " + r + " row" + ( r != 1 ? "s" : "" ) + ".";
            }
        }
        catch ( Exception e ) {
            return Bundle.BUNDLE.getString( "messagePrompt" ) + " " + new ColoredString( Bundle.BUNDLE.getString( "error" ), AnsiColor.ERROR_COLOR ) + " " + e.getMessage();
        }
        finally {
            closeResources( conn, stmt, rs );
        }
    }

    /**
     * Executes a data definition command, such as alter table, etc.
     *
     * @param sql
     * @return     Description of the Returned Value
     */
    public String doDDL( String sql ) {
        Connection conn = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;

        try {
            conn = getConnection();
            stmt = conn.prepareStatement( sql );
            stmt.executeUpdate();
            return Bundle.BUNDLE.getString( "messagePrompt" ) + " Command complete.";
        }
        catch ( Exception e ) {
            return Bundle.BUNDLE.getString( "messagePrompt" ) + " " + new ColoredString( Bundle.BUNDLE.getString( "error" ), AnsiColor.ERROR_COLOR ) + " " + e.getMessage();
        }
        finally {
            closeResources( conn, stmt, rs );
        }
    }

    /**
     * Converts a ResultSet to a TableArray. The left most column of the
     * TableArray is column 1, just like in a ResultSet. All values are stored
     * in the TableArray as Strings. The top row of the returned TableArray will
     * contain the column names.<p>
     * BLOBs, CLOBs, and VARBINARY data are displayed as BLOB, CLOB, and VARBINARY,
     * respectively, rather than print out a bunch of meaningless binary data to
     * the screen.  If you need the raw binary data, use the \raw command in
     * DbConsole.
     *
     * @param rs                the result set to copy into a TableArray
     * @return                  a TableArray containing the result set data as
     *      Strings. The top row of the TableArray will contain the column
     *      names.
     * @exception SQLException  Description of Exception
     */
    public TableArray resultSetToTableArray( ResultSet rs ) throws SQLException {
        int y = 0;
        table = new TableArray();
        boolean show_row_col = YES.equals( System.getProperty( SHOW_ROW_COLUMN ) );

        // set up columns and headers
        if ( show_row_col )
            table.put( 1, y, new ColoredString( "ROW", AnsiColor.CYAN ) );
        ResultSetMetaData md = rs.getMetaData();
        int column_count = md.getColumnCount();
        for ( int x = 1; x <= column_count; x++ ) {
            if ( !CAN_RUN )
                break;
            ColoredString data = new ColoredString( md.getColumnName( x ).toUpperCase(), AnsiColor.CYAN );
            table.put( x + ( show_row_col ? 1 : 0 ), y, data );
        }
        ++y;

        while ( rs.next() ) {
            if ( !CAN_RUN ) {
                break;
            }
            String data = null;
            for ( int x = 1; x <= column_count; x++ ) {
                if ( show_row_col )
                    table.put( 1, y, new ColoredString( String.valueOf( y ), AnsiColor.CYAN ) );
                switch ( md.getColumnType( x ) ) {
                    case Types.LONGVARBINARY:
                    case Types.VARBINARY:
                        data = "VARBINARY";
                        break;
                    case Types.BLOB:
                        data = "BLOB";
                        break;
                    case Types.CLOB:
                        Clob clob = rs.getClob( x );
                        if ( clob != null ) {
                            int length = ( new Long( clob.length() ) ).intValue();
                            data = clob.getSubString( 1L, length );
                        }
                        else {
                            data = "CLOB";
                        }
                        break;
                    case Types.TIMESTAMP:
                        Timestamp t = rs.getTimestamp(x);
                        data = t == null ? "" : t.toString();
                        break;
                    default:
                        Object o = rs.getObject( x );
                        data = o == null ? "" : o.toString(); //.trim();
                        break;
                }
                table.put( x + ( show_row_col ? 1 : 0 ), y, data );
            }
            ++y;
        }
        if ( getLandscapeMode() )
            return convertToLandscape();
        return table;
    }

    /**
     * Returns a raw string containing the contents of the result set.  No attempt
     * it made to format the contents other than a line separator is inserted in
     * between rows of data.  Column data is not separated.  This is most useful
     * when querying for a single binary value and you want to examine the raw
     * bytes.
     *
     * @param rs
     * @return                  Description of the Returned Value
     * @exception SQLException  Description of Exception
     */
    public String resultSetToRaw( ResultSet rs ) throws SQLException {
        int y = 0;
        StringWriter sw = new StringWriter();
        ResultSetMetaData md = rs.getMetaData();
        int column_count = md.getColumnCount();

        while ( rs.next() ) {
            if ( !CAN_RUN ) {
                break;
            }
            String data = null;
            for ( int x = 1; x <= column_count; x++ ) {
                try {
                    switch ( md.getColumnType( x ) ) {
                        case Types.LONGVARBINARY:
                        case Types.VARBINARY:
                        case Types.BINARY:
                        case Types.BLOB:
                            InputStream is = rs.getBinaryStream( x );
                            FileUtilities.copyToWriter( new InputStreamReader( is ), sw );
                            break;
                        case Types.CLOB:
                            is = rs.getAsciiStream( x );
                            FileUtilities.copyToWriter( new InputStreamReader( is ), sw );
                            break;
                        default:
                            Object o = rs.getObject( x );
                            data = o == null ? "" : o.toString();
                            FileUtilities.copyToWriter( new StringReader( data ), sw );
                            break;
                    }
                }
            catch ( Exception e ) {}
            }
            try {
                sw.write( LS );
            }
            catch ( Exception e ) {
                e.printStackTrace();
            }
            ++y;
        }
        return sw.toString();
    }

    /**
     * Converts the given TableArray to a formatted string. Assumes the first
     * row of the table is a header row containing the names of the columns.
     * Here's an example of the output: <br>
     * <pre>
     *  |TEAM_ID|TEAM_NAME |CITY         |
     *  +-------+----------+-------------+
     *  |1      |Ravens    |Baltimore    |
     *  |2      |Bills     |Buffalo      |
     *  |3      |Bengals   |Cincinnati   |
     *  |4      |Browns    |Cleveland    |
     * </pre>
     * Adjusted so that first column is always "ROW" and the value will be the
     * row number.  This can be used by the "show row" command to better see
     * a single row.
     *
     * @param table
     * @return       Description of the Returned Value
     */
    public String tableArrayToFormattedString( TableArray table ) {

        // calculate column widths
        int[] col_widths = new int[ table.getWidth() ];
        for ( int i = 1; i <= table.getWidth(); i++ ) {
            if ( !CAN_RUN ) {
                break;
            }
            List column = table.getColumn( i );
            int max_width = 0;
            for ( Iterator it = column.iterator(); it.hasNext(); ) {
                Object o = it.next();
                CharSequence value = o == null ? "" : ( CharSequence ) o;
                if ( value.length() > max_width ) {
                    max_width = value.length();
                }
            }
            col_widths[ i - 1 ] = max_width;
        }

        StringBuffer sb = new StringBuffer();
        String coloredPipe = new ColoredPipe().toString();
        String coloredPlus = new ColoredPlus().toString();

        // build the header
        Object[] row = table.getRow( 0 ).toArray();
        sb.append( coloredPipe );
        for ( int i = 0; i < row.length; i++ ) {
            if ( !CAN_RUN ) {
                break;
            }
            CharSequence value = row[ i ] == null ? "" : ( CharSequence ) row[ i ];
            sb.append( value.toString() );
            for ( int j = value.length(); j < col_widths[ i ]; j++ )
                sb.append( ' ' );
            sb.append( coloredPipe );
        }
        sb.append( LS );

        // build the horizontal line to separate the header from the data
        sb.append( coloredPlus );
        for ( int i = 0; i < col_widths.length; i++ ) {
            if ( !CAN_RUN )
                break;
            for ( int j = 0; j < col_widths[ i ]; j++ )
                sb.append( '-' );
            sb.append( coloredPlus );
        }
        sb.append( LS );

        // add the data
        int y = 1;
        for ( ; y < table.getRowCount(); y++ ) {
            if ( !CAN_RUN ) {
                break;
            }
            row = table.getRow( y ).toArray();
            sb.append( coloredPipe );
            for ( int i = 0; i < row.length; i++ ) {
                CharSequence value = row[ i ] == null ? "" : ( CharSequence ) row[ i ];
                sb.append( value.toString() );
                for ( int j = value.length(); j < col_widths[ i ]; j++ )
                    sb.append( ' ' );
                sb.append( coloredPipe );
            }
            sb.append( LS );
        }

        sb.append( LS ).append( "(" ).append( y - 1 ).append( " row" ).append( y - 1 != 1 ? "s" : "" ).append( ")" ).append( LS );
        return sb.toString();
    }

    /**
     * Convenience method to close database connection resources.
     *
     * @param conn
     * @param stmt
     * @param rs
     */
    private void closeResources( Connection conn, Statement stmt, ResultSet rs ) {
        if ( autoCommitMode ) {
            try {
                conn.close();
            }
            catch ( Exception e ) {}
            try {
                stmt.close();
            }
            catch ( Exception e ) {}
            try {
                rs.close();
            }
            catch ( Exception e ) {}
        }
    }

    /**
     * @param conn
     * @param rs
     */
    private void closeResources( Connection conn, ResultSet rs ) {
        if ( autoCommitMode ) {
            try {
                conn.close();
            }
            catch ( Exception e ) {}
            try {
                rs.close();
            }
            catch ( Exception e ) {}
        }
    }

    private void closeConnection( Connection conn ) {
        try {
            conn.close();
        }
        catch ( Exception e ) {}
    }

}
