/*
Copyright (c) Dale Anson, 2004
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote products
derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package ise.dbconsole.command;

import java.io.*;
import java.util.*;
import java.util.prefs.*;

import ise.dbconsole.*;
import ise.dbconsole.to.*;
import ise.library.*;
import ise.library.ascii.*;

/**
 * @version   $Revision: 18 $
 */
public class EditDbInfoCommand extends Command {

    /**
     * Execute the command. A common implementation is to delegate to <code>
     * execute(Object)</code> and pass null.
     *
     * @return   Description of the Returned Value
     */
    public String execute() {
        return execute(null);
    }

    /**
     * Execute the command.
     *
     * @param data  data for the command.
     * @return      Description of the Returned Value
     */
    public String execute(Object data) {
        DbConsole console = DbConsole.getInstance();
        PrintStream out = console.getOut();

        // pick db
        // choice: edit info, edit users
        // edit info choice:
        //    choice to edit name, url, driver
        //    save changes
        // edit users choice:
        //    choose user, add new user
        //    choose user choice:
        //       edit, delete
        //    save changes

        List db_configs = DbConsoleUtil.getDbConfigurations();
        if (db_configs == null || db_configs.size() == 0) {
            return MessageBox.box("Configuration Error", "There are currently no configured databases.");
        }

        // have the user choose a db to edit
        ChoiceInput choiceField = new ChoiceInput(db_configs);
        choiceField.setResourceBundle(Bundle.BUNDLE);
        choiceField.setIntro("Select database configuration to edit" + ": ");
        Object o = choiceField.getChoice();
        DbInfo dbinfo;
        if (o != ChoiceInput.USER_CANCELLED) {
            dbinfo = (DbInfo) o;
        }
        else {
            // user selected cancel, so just return without doing anything else
            return null;
        }

        // show the db configuration
        out.println(MessageBox.box("Current settings for this configuration:", dbinfo.toVerboseString()));

        // delete or edit?
        choiceField = new ChoiceInput();
        Choice editChoice = new Choice("Edit");
        Choice deleteChoice = new Choice("Delete");
        choiceField.addChoice(editChoice);
        choiceField.addChoice(deleteChoice);
        choiceField.setShowCancel(true);
        Object todo = choiceField.getChoice();
        if (todo == null || todo == ChoiceInput.USER_CANCELLED) {
            return null;
        }
        else if (todo == deleteChoice) {
            ChoiceInput confirm = ChoiceInput.createYesNoChoice("Confirm delete of " + dbinfo.getName());
            if (confirm.getChoice() == ChoiceInput.YES_CHOICE) {
                DbConsoleUtil.removeDbConfig(dbinfo);   
            }
            return null;
        }
        else {
            // let the user edit the configuration
            String name = null;
            do {
                Input input = new TextInput("Change name:");
                input.setDefaultValue(dbinfo.getName());
                name = input.getInput();
                if (name == null || name.length() == 0) {
                    name = null;
                    out.println("Invalid entry.");
                }
            } while (name == null);
            dbinfo.setName(name);

            String url = null;
            do {
                Input input = new TextInput("Change URL:");
                input.setDefaultValue(dbinfo.getUrl());
                url = input.getInput();
                if (url == null || url.length() == 0) {
                    url = null;
                    out.println("Invalid entry.");
                }
            } while (url == null);
            dbinfo.setUrl(url);

            String driver = null;
            do {
                Input input = new TextInput("Change JDBC driver:");
                input.setDefaultValue(dbinfo.getDriver());
                driver = input.getInput();
                if (driver == null || driver.length() == 0) {
                    driver = null;
                    out.println("Invalid entry.");
                }
            } while (driver == null);
            dbinfo.setDriver(driver);

            // see if we should edit the users for this configuration
            ChoiceInput ci = ChoiceInput.createYesNoChoice("Do you want to add/edit/delete users for this configuration?");
            Object choice = ci.getChoice();
            if (choice == ChoiceInput.YES_CHOICE) {
                List users = dbinfo.getUsers();
                Choice add_user = new Choice("Add new user");
                users.add(add_user);
                choiceField = new ChoiceInput(users);
                choiceField.setResourceBundle(Bundle.BUNDLE);
                choiceField.setIntro("Select username to edit" + ": ");
                o = choiceField.getChoice();
                if (o == ChoiceInput.USER_CANCELLED) {
                    // do nothing
                    return null;
                }
                else if (o == add_user) {
                    out.println("Add new user:");
                    User new_user = DbConsoleUtil.createUser();
                    if (new_user != null)
                        dbinfo.addUser(new_user);
                }
                else {
                    User user = (User) o;
                    String username = user.getUsername();
                    // option to change password or delete user
                    Choice pw = new Choice("Change password for " + username);
                    Choice delete = new Choice("Delete " + username);
                    ci = new ChoiceInput();
                    ci.addChoice(pw);
                    ci.addChoice(delete);
                    ci.setIntro("Select option for " + username);
                    Choice c = (Choice) ci.getChoice();
                    if (c == pw) {
                        ChangePasswordCommand cpc = new ChangePasswordCommand();
                        String new_password = cpc.execute(user);
                        if (new_password != null) {
                            dbinfo.removeUser(user);
                            user.setPassword(new_password);
                            dbinfo.addUser(user);
                        }
                    }
                    else {
                        ci = ChoiceInput.createYesNoChoice("Are you sure you want to delete " + username + "?");
                        choice = ci.getChoice();
                        if (choice == ChoiceInput.YES_CHOICE) {
                            dbinfo.removeUser(user);
                        }
                    }
                }
            }
            DbConsoleUtil.storeDbInfo(dbinfo);
        }
        return null;
    }

    /**
     * Should DbConsole show the spinner? Most commands can leave this as is, it
     * usually doesn't hurt anything to show the spinner.
     *
     * @return   This implementation returns false.
     */
    public boolean showSpinner() {
        return false;
    }
}

