/*
Copyright (c) Dale Anson, 2004
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote products
derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package ise.dbconsole.command;

import java.io.*;
import java.rmi.server.UID;
import java.util.*;

import ise.dbconsole.*;
import ise.dbconsole.to.*;
import ise.library.*;
import ise.library.ascii.*;

/**
 * Sets up the DAO. The DAO for this application is treated as a singleton,
 * although it doesn't follow the singleton pattern.
 *
 * @version   $Revision: 16 $
 */
public class SetupDAOCommand2 extends Command implements Constants {
    private DbConsole console = null;
    private InputStream in = null;
    private PrintStream out = null;

    private final static String DBINFO = "dbinfo";


    /**
     * Execute the command. A common implementation is to delegate to <code>
     * execute(Object)</code> and pass null.
     *
     * @return   Description of the Returned Value
     */
    public String execute() {
        return execute(null);
    }

    /**
     * Execute the command.
     *
     * @param data  data for the command.
     * @return      Description of the Returned Value
     */
    public String execute(Object data) {
        console = DbConsole.getInstance();
        in = console.getIn();
        out = console.getOut();

        // store the current db info, just in case the user changes to a
        // different db
        saveCurrentDbInfo();

        setupDAO();
        return null;
    }

    /**
     * Tell DbConsole to NOT show a spinner.
     *
     * @return   false
     */
    public boolean showSpinner() {
        return false;
    }

    private void saveCurrentDbInfo() {
        DbConsoleDAO dao = console.getDAO();
        if (dao != null) {
            DbInfo info = dao.getCurrentDbInfo();
            DbConsoleUtil.storeDbInfo(info);
        }
    }

    /**
     * Load the data access object. Gets the username and password from the
     * command prompt to pass to the DAO.
     */
    private void setupDAO() {
        showCurrentConnection();

        List db_configs = DbConsoleUtil.getDbConfigurations();
        if (db_configs == null || db_configs.size() == 0) {
            DbInfo dbinfo = createDbConfiguration();
            db_configs = new ArrayList();
            db_configs.add(dbinfo);
        }
        db_configs.add(Bundle.BUNDLE.getString("other"));

        DbInfo default_config = DbConsoleUtil.getDefaultDbConfiguration();

        // fetch the user's choice
        ChoiceInput choiceField = new ChoiceInput(db_configs);
        choiceField.setResourceBundle(Bundle.BUNDLE);
        choiceField.setIntro(Bundle.BUNDLE.getString("selectUrlAndDriver") + ": ");
        if (default_config != null)
            choiceField.setDefaultChoice(default_config);
        Object o = choiceField.getChoice();
        DbInfo dbinfo;
        if (o != ChoiceInput.USER_CANCELLED && !Bundle.BUNDLE.getString("other").equals(o)) {
            dbinfo = (DbInfo) o;
        }
        else if (o == ChoiceInput.USER_CANCELLED) {
            // user selected cancel, so just return without doing anything else
            return;
        }
        else {
            // user selected other, so get a new configuration
            dbinfo = createDbConfiguration();
        }

        if (dbinfo.getUserCount() == 0)
            addDbLogin(dbinfo);
        if (dbinfo.getUserCount() == 0) {
            out.println(Bundle.BUNDLE.getString("noUsersFor"));
            out.println(dbinfo.toString());
            out.println(Bundle.BUNDLE.getString("notConnected"));
        }

        // pick the username to use to connect to the database
        List users = dbinfo.getUsers();
        users.add(Bundle.BUNDLE.getString("other"));
        choiceField = new ChoiceInput(users);
        choiceField.setResourceBundle(Bundle.BUNDLE);
        choiceField.setIntro(Bundle.BUNDLE.getString("selectUsernameForConnection") + ": ");
        if (dbinfo.getUser() != null)
            choiceField.setDefaultChoice(dbinfo.getUser());
        o = choiceField.getChoice();
        if (o != ChoiceInput.USER_CANCELLED && !Bundle.BUNDLE.getString("other").equals(o)) {
            dbinfo.setUser((User)o);
        }
        else if (o == ChoiceInput.USER_CANCELLED) {
            // user selected cancel, so just return without doing anything else
            return;
        }
        else {
            // user selected other, so add a new login
            addDbLogin(dbinfo);
        }
        DbConsoleUtil.storeDbInfo(dbinfo);
        DbConsoleUtil.setDefaultDbConfiguration(dbinfo);

        User user = dbinfo.getUser();
        String ep = user.getPassword();
        if (ep == null || ep.length() == 0) {
            out.println("No password found for " + user.getUsername());
            Input input = new PasswordInput( DbConsole.prompt + "Password: " );
            user.setPassword(input.getInput());
        }
        createDAO(dbinfo);

        console.setSavedQueries(dbinfo.getSavedQueries());
    }

    /**
     * Description of the Method
     *
     * @param dbinfo
     */
    private void createDAO(DbInfo dbinfo) {
        // load the db driver and url, set up the DAO
        try {
            out.println();
            out.print(Bundle.BUNDLE.getString("connecting"));
            console.setDAO(new DbConsoleDAO(dbinfo));

            // run the simple db info command to verify the connection and
            // show the user where they are connected
            DbInfoCommand c = new DbInfoCommand();
            String result = c.execute();
            out.println(result);
            console.prompt = console.getDAO().getCurrentDbInfo().getName() + "> ";
        }
        catch (ClassNotFoundException cnfe) {
            // this won't be good
            cnfe.printStackTrace();
            out.println(cnfe.getMessage());
            out.println(Bundle.BUNDLE.getString("looksLikeTheDriverWontLoad"));
            System.exit(1);
        }
        catch (Exception e) {
            // this won't be good either
            e.printStackTrace();
            out.println(e.getMessage());
            out.println(Bundle.BUNDLE.getString("looksLikeThereWasAProblem"));
            System.exit(1);
        }
    }

    /**
     * Description of the Method
     *
     * @return   Description of the Returned Value
     */
    private DbInfo createDbConfiguration() {
        // get user input

        out.println(Bundle.BUNDLE.getString("createNewDatabaseConfiguration"));
        Input input = new TextInput(Bundle.BUNDLE.getString("enterNameForThisConfiguration") + ":");
        String name = input.getInput();
        input = new TextInput(Bundle.BUNDLE.getString("enterDatabaseUrl") + ":");
        String url = input.getInput();
        input = new TextInput(Bundle.BUNDLE.getString("enterDriverClassname") + ":");
        String driver = input.getInput();

        // create the dbinfo
        DbInfo info = new DbInfo();
        info.setId(String.valueOf(new Date().getTime()));
        info.setName(name);
        info.setUrl(url);
        info.setDriver(driver);

        // add a user
        addDbLogin(info);

        // store the dbinfo
        DbConsoleUtil.storeDbInfo(info);

        return info;
    }

    /**
     * adds a user to the given dbinfo
     *
     * @param c  a DbUrlChoice, which is a transfer object for the database url
     *      and driver classname
     */
    private void addDbLogin(DbInfo c) {
        out.println();
        out.println(Bundle.BUNDLE.getString("needUserNameAndPassword") + ":");
        out.println("\t" + c.getUrl());
        out.println(Bundle.BUNDLE.getString("usingDriverClass") + ":");
        out.println("\t" + c.getDriver());
        out.println();

        String username;
        String password;
        try {
            User user = new User();
            Input input = new TextInput(DbConsole.prompt + Bundle.BUNDLE.getString("enterUsername") + ":");
            username = input.getInput();
            user.setUsername(username);
            ChangePasswordCommand cpc = new ChangePasswordCommand();
            cpc.execute(user);
            c.addUser(user);
            c.setUser(user);
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }

    /** show which database currently connected to */
    private void showCurrentConnection() {
        DbConsoleDAO dao = console.getDAO();
        if (dao != null) {
            DbInfo info = dao.getCurrentDbInfo();
            out.println(Bundle.BUNDLE.getString("currentConnection") + " " + info.toString());
        }
    }

}

