/*
Copyright (c) Dale Anson, 2004
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
  3. The name of the author may not be used to endorse or promote products
  derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package ise.dbconsole.to;

import java.util.*;
import ise.dbconsole.*;
import ise.library.*;

/**
 * Transfer object for database information.
 *
 * @version   $Revision: 15 $
 */
public class DbInfo implements Comparable {
    private String dbname;
    private String dburl;
    private String dbdriver;
    private String id;
    private ArrayList dbusers;
    private User user;
    private ArrayList recentQueries = new ArrayList();

    public String toString() {
        return dbname;
    }

    /**
     * Shows the database name, url, and driver in a form suitable for display.
     *
     * @return   Description of the Returned Value
     */
    public String toVerboseString() {
        String ls = System.getProperty( "line.separator" );
        StringBuffer sb = new StringBuffer();
        sb.append( Bundle.BUNDLE.getString("database")).append(":").append( dbname ).append( ls );
        sb.append( Bundle.BUNDLE.getString("url") ).append(":").append( dburl ).append( ls );
        sb.append( Bundle.BUNDLE.getString("driver") ).append(":").append( dbdriver ).append( ls );
        sb.append( Bundle.BUNDLE.getString("users") ).append(":").append( ls );
        for ( Iterator it = getUsers().iterator(); it.hasNext(); ) {
            sb.append( "\t" ).append( ( ( User ) it.next() ).getUsername() ).append( ls );
        }
        return sb.toString();
    }

    /**
     * Used to sort a list of DbInfos, sorting is done by the database name.
     *
     * @param o
     * @return   Description of the Returned Value
     */
    public int compareTo( Object o ) {
        if ( o == null )
            return -1;
        return dbname.compareTo( ( ( DbInfo ) o ).getName() );
    }

    /**
     * Sets the id of this DbInfo, this is for internal use by DbConsole for
     * storing this DbInfo.
     *
     * @param id  The new id value
     */
    public void setId( String id ) {
        this.id = id;
    }

    /**
     * Gets the id of this database, this is for internal use by DbConsole for
     * storing this DbInfo.
     *
     * @return   The id value
     */
    public String getId() {
        return id;
    }

    /**
     * Sets the common name for this database.
     *
     * @param name  The new name value
     */
    public void setName( String name ) {
        if ( name == null || name.length() == 0 )
            throw new IllegalArgumentException( Bundle.BUNDLE.getString("name") + " " + Bundle.BUNDLE.getString("cannotBeNullOrEmpty") );
        dbname = name;
    }

    /**
     * Gets the common name for this database.
     *
     * @return   The name value
     */
    public String getName() {
        return dbname;
    }

    /**
     * Sets the url for this database.
     *
     * @param url  The new url value
     */
    public void setUrl( String url ) {
        if ( url == null || url.length() == 0 )
            throw new IllegalArgumentException( Bundle.BUNDLE.getString(url) + " " + Bundle.BUNDLE.getString("cannotBeNullOrEmpty") );
        dburl = url;
    }

    /**
     * Gets the url for this database.
     *
     * @return   The url value
     */
    public String getUrl() {
        return dburl;
    }

    /**
     * Sets the classname of the driver for this database.
     *
     * @param driver  The new driver value
     */
    public void setDriver( String driver ) {
        if ( driver == null || driver.length() == 0 )
            throw new IllegalArgumentException( Bundle.BUNDLE.getString("driver") + " " + Bundle.BUNDLE.getString("cannotBeNullOrEmpty") );
        dbdriver = driver;
    }

    /**
     * Gets the classname of the database driver for this database.
     *
     * @return   The driver value
     */
    public String getDriver() {
        return dbdriver;
    }

    /**
     * Adds a known user to the list of users for this database
     *
     * @param user  The feature to be added to the User attribute
     */
    public void addUser( User user ) {
        if ( dbusers == null )
            dbusers = new ArrayList();
        dbusers.add( user );
        Collections.sort( dbusers );
    }

    public void removeUser( User user ) {
        dbusers.remove( user );
    }

    /**
     * Gets an iterator of all the users that have been set up for this
     * database.
     *
     * @return   The users value
     */
    public List getUsers() {
        if ( dbusers == null )
            dbusers = new ArrayList();
        List copy = new ArrayList( dbusers );
        return copy;
    }

    /**
     * Gets the userCount attribute of the DbInfo object
     *
     * @return   The userCount value
     */
    public int getUserCount() {
        if ( dbusers == null )
            return 0;
        return dbusers.size();
    }

    /**
     * Gets username and password for the user currently set to use to connect
     * to the database.
     *
     * @return   The user value
     */
    public Properties getUserProperties() {
        if ( user == null )
            return null;
        try {
            Properties p = new Properties();
            p.put( "user", user.getUsername() );
            p.put( "password", user.getPlainPassword() );
            return p;
        }
        catch ( Exception e ) {
            e.printStackTrace();
            return null;
        }
    }

    public User getUser() {
        return user;
    }

    /**
     * Sets the specific user to use to connect to the database.
     *
     * @param user  The new user value
     */
    public void setUser( User user ) {
        this.user = user;
    }

    public void setUser(String name) {
        for (Iterator it = dbusers.iterator(); it.hasNext(); ) {
            User user = (User)it.next();
            if (user.getUsername().equals(name)) {
                setUser(user);
                return;
            }
        }
    }

    public void setSavedQueries(ArrayList rs) {
        recentQueries = rs;
    }

    public ArrayList getSavedQueries() {
        return recentQueries;
    }
}

