/*
Copyright (c) Dale Anson, 2004
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
  3. The name of the author may not be used to endorse or promote products
  derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package ise.library.ascii;

/**
 * An implementation of the standard ansi color codes. I borrowed the color
 * definitions from Ant's AnsiColorLogger.
 *
 * @author    Dale Anson
 * @version   $Revision: 10 $
 */
public class AnsiColor {

    // most of these should be self-explanatory...
    public final static int ATTR_NORMAL = 0;
    public final static int ATTR_BRIGHT = 1;
    public final static int ATTR_DIM = 2;
    public final static int ATTR_UNDERLINE = 3;
    public final static int ATTR_BLINK = 5;
    public final static int ATTR_REVERSE = 7;
    public final static int ATTR_HIDDEN = 8;

    public final static int FG_BLACK = 30;
    public final static int FG_RED = 31;
    public final static int FG_GREEN = 32;
    public final static int FG_YELLOW = 33;
    public final static int FG_BLUE = 34;
    public final static int FG_MAGENTA = 35;
    public final static int FG_CYAN = 36;
    public final static int FG_WHITE = 37;

    public final static int BG_BLACK = 40;
    public final static int BG_RED = 41;
    public final static int BG_GREEN = 42;
    public final static int BG_YELLOW = 44;
    public final static int BG_BLUE = 44;
    public final static int BG_MAGENTA = 45;
    public final static int BG_CYAN = 46;
    public final static int BG_WHITE = 47;

    private final static String PREFIX = "\u001b[";
    private final static String SUFFIX = "m";
    private final static char SEPARATOR = ';';
    public final static String END_COLOR = PREFIX + SUFFIX;

    /* these 2 'escaped' codes are convenient for when the color codes need to
    be stripped from a colorized string.  These codes can be used directly
    in String.replaceAll. */
    private final static String ESCAPED_PREFIX = "\u001b\\[";
    public final static String ESCAPED_END_COLOR = ESCAPED_PREFIX + SUFFIX;

    // default color is cyan
    private int colorCode = FG_CYAN;

    private int attributes = ATTR_DIM;

    // some predefined colors
    public final static AnsiColor ERROR_COLOR = new AnsiColor(FG_RED, ATTR_BRIGHT);
    public final static AnsiColor WARN_COLOR = new AnsiColor(FG_MAGENTA, ATTR_NORMAL);

    // standard foreground colors
    public final static AnsiColor CYAN = new AnsiColor(FG_CYAN);
    public final static AnsiColor BLACK = new AnsiColor(FG_BLACK);
    public final static AnsiColor RED = new AnsiColor(FG_RED);
    public final static AnsiColor GREEN = new AnsiColor(FG_GREEN);
    public final static AnsiColor YELLOW = new AnsiColor(FG_YELLOW);
    public final static AnsiColor BLUE = new AnsiColor(FG_BLUE);
    public final static AnsiColor MAGENTA = new AnsiColor(FG_MAGENTA);
    public final static AnsiColor WHITE = new AnsiColor(FG_WHITE);


    /**
     * Constructor for AnsiColor
     *
     * @param color_code
     */
    private AnsiColor(int color_code) {
        colorCode = color_code;
    }

    private AnsiColor(int color_code, int attributes) {
        colorCode = color_code;
        this.attributes = attributes;
    }

    /**
     * Description of the Method
     *
     * @param foreground
     * @return            Description of the Returned Value
     */
    public static AnsiColor createColor(int foreground) {
        switch (foreground) {
            case FG_BLACK:
            case FG_RED:
            case FG_GREEN:
            case FG_YELLOW:
            case FG_BLUE:
            case FG_MAGENTA:
            case FG_CYAN:
            case FG_WHITE:
                return new AnsiColor(foreground);
            default:
                return CYAN;
        }
    }

    /**
     * Description of the Method
     *
     * @param foreground One of black, red, green, yellow, blue, magenta, or white.
     * If null or any other string, will return cyan.
     * @return  An AnsiColor corresponding to the given foreground color.
     */
    public static AnsiColor createColor(String foreground) {
        if (foreground == null)
            return CYAN;
        foreground = foreground.toLowerCase();
        if (foreground.equals("black"))
            return BLACK;
        if (foreground.equals("red"))
            return RED;
        if (foreground.equals("green"))
            return GREEN;
        if (foreground.equals("yellow"))
            return YELLOW;
        if (foreground.equals("blue"))
            return BLUE;
        if (foreground.equals("magenta"))
            return MAGENTA;
        if (foreground.equals("white"))
            return WHITE;
        return CYAN;
    }

    /**
     * Description of the Method
     *
     * @return   Description of the Returned Value
     */
    public String toString() {
        return PREFIX + attributes + SEPARATOR + colorCode + SUFFIX;
    }

    public String toEscapedString() {
        return ESCAPED_PREFIX + attributes + SEPARATOR + colorCode + SUFFIX;
    }
}

