/*
Copyright (c) Dale Anson, 2004
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
  3. The name of the author may not be used to endorse or promote products
  derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package ise.library.ascii;

import java.io.*;
import java.util.*;
import ise.dbconsole.*;

/**
 * ChoiceInput. A generic command-line chooser. Displays a numbered list of
 * items and returns the selected item from the list. <br>
 * InputStream = System.in<br>
 * OutputStream = System.out<br>
 * List = choices to display<br>
 * default = default choice<br>
 * showCancel = show a "Cancel" choice<br>
 * return = list object<br>
 *
 *
 * @version   $Revision: 10 $
 */
public class ChoiceInput {

    public final static Choice USER_CANCELLED = new Choice(Bundle.BUNDLE.getString("cancel"));
    public final static Choice YES_CHOICE = new Choice( Bundle.BUNDLE.getString("yes") );
    public final static Choice NO_CHOICE = new Choice( Bundle.BUNDLE.getString("no") );

    public static String LS = System.getProperty( "line.separator" );

    private String intro = "Choose:";
    private static ResourceBundle bundle = null;

    private List choices = null;
    private boolean showCancel = true;

    private Object defaultChoice = null;

    /** Constructor for ChoiceInput */
    public ChoiceInput() {
        this( null, true );
    }

    /**
     * Constructor for ChoiceInput
     *
     * @param choices
     */
    public ChoiceInput( Collection choices ) {
        this( choices, true );
    }

    /**
     * Constructor for ChoiceInput
     *
     * @param choices     the list of choices
     * @param showCancel  whether a cancel option should be shown.
     */
    public ChoiceInput( Collection choices, boolean showCancel ) {
        this.choices = choices == null ? new ArrayList() : new ArrayList(choices);
        this.showCancel = showCancel;
        bundle = getDefaultResourceBundle();
    }
    
    /**
     * This string will be shown above the list of choices.
     *
     * @param intro  The new intro value
     */
    public void setIntro( String intro ) {
        this.intro = intro;
    }

    /**
     * The string that will be shown above the list of choices.
     *
     * @return   The intro value
     */
    public String getIntro() {
        return intro;
    }

    /**
     * Sets the resource bundle for display strings.
     *
     * @param bundle  The new resourceBundle value
     * @see           getDefaultResourceBundle for valid keys
     */
    public void setResourceBundle( ResourceBundle bundle ) {
        this.bundle = bundle;
    }

    /**
     * Gets the resourceBundle attribute of the ChoiceInput object
     *
     * @return   The resourceBundle value
     * @see      getDefaultResourceBundle for valid keys
     */
    public ResourceBundle getResourceBundle() {
        return bundle;
    }

    /**
     * Gets the defaultResourceBundle for the ChoiceInput, here are the valid
     * keys for any reource bundle and default values:
     * <ul>
     *   <li> choiceFieldPrompt=Choose:</li>
     *   <li> choiceFieldCancel=Cancel</li>
     *   <li> choiceFieldCancelChar=x</li>
     *   <li> choiceFieldInvalidChoice=Invalid choice, try again.</li>
     *   <li> choiceFieldProblem=Error.</li>
     * </ul>
     *
     *
     * @return   The defaultResourceBundle value
     */
    public ResourceBundle getDefaultResourceBundle() {
        try {
            String props =
                "choiceFieldPrompt=Choose:\n" +
                "choiceFieldCancel=Cancel\n" +
                "choiceFieldCancelChar=x\n" +
                "choiceFieldInvalidChoice=Invalid choice, try again.\n" +
                "choiceFieldProblem=Error.\n" +
                "yes=Yes\n" +
                "no=No\n";
            return new PropertyResourceBundle( new ByteArrayInputStream( props.getBytes() ) );
        }
        catch ( IOException ioe ) {
            throw new RuntimeException( ioe.getMessage() );
        }
    }

    /**
     * Sets the choices.
     *
     * @param choices  The new choices value
     */
    public void setChoices( Collection choices ) {
        this.choices = new ArrayList(choices);
    }

    /**
     * Gets the choices attribute of the ChoiceInput object
     *
     * @return   The choices value
     */
    public List getChoices() {
        return choices;
    }

    public void addChoice( Object o ) {
        if ( choices == null )
            choices = new ArrayList();
        choices.add( o );
    }

    public void removeChoice( Object o ) {
        if ( choices != null )
            choices.remove( o );
    }

    /**
     * Gets the choice attribute of the ChoiceInput object
     *
     * @return   The choice value
     */
    public Object getChoice() {
        if ( choices == null || choices.size() == 0 )
            return defaultChoice;
        do {

            StringBuffer sb = new StringBuffer();

            // get choices
            for ( int i = 0; i < choices.size(); i++ ) {
                String index = new ColoredString(String.valueOf(i + 1) + ") ", AnsiColor.CYAN).toString();
                sb.append( index ).append( choices.get( i ).toString() ).append( LS );
            }

            // maybe add the cancel
            if ( showCancel ) {
                String cancelChoice = new ColoredString(bundle.getString( "choiceFieldCancelChar" ) + ") ", AnsiColor.CYAN).toString();
                sb.append( cancelChoice + bundle.getString( "choiceFieldCancel" ) );
            }

            System.out.println( MessageBox.box( intro, sb.toString() ) );

            // show prompt
            String prompt = bundle.getString( "choiceFieldPrompt" ) + " ";
            if ( defaultChoice != null )
                prompt += "(" + defaultChoice.toString() + ")";

            try {
                Input infield = new TextInput( prompt );
                String input = infield.getInput();

                // check for default choice
                if ( ( input == null || input.length() == 0 ) && defaultChoice != null )
                    return defaultChoice;

                if ( input == null || input.length() == 0 ) {
                    System.out.println( bundle.getString( "choiceFieldInvalidChoice" ) );
                    continue;
                }

                // check for cancel
                if ( input.startsWith( bundle.getString( "choiceFieldCancelChar" ) ) )
                    return USER_CANCELLED;

                int index = Integer.parseInt( input ) - 1;
                return choices.get( index );
            }
            catch ( NumberFormatException nfe ) {
                System.out.println( bundle.getString( "choiceFieldInvalidChoice" ) );
                continue;
            }
            catch ( IndexOutOfBoundsException oobe) {
                System.out.println( bundle.getString( "choiceFieldInvalidChoice" ) );
                continue;
            }
        }
        while ( true );
    }

    /**
     * Sets the defaultChoice attribute of the ChoiceInput object
     *
     * @param i  The new defaultChoice value
     */
    public void setDefaultChoice( Object i ) {
        defaultChoice = i;
    }

    /**
     * Gets the defaultChoice attribute of the ChoiceInput object
     *
     * @return   The defaultChoice value
     */
    public Object getDefaultChoice() {
        return defaultChoice;
    }

    /**
     * Sets the showCancel attribute of the ChoiceInput object
     *
     * @param b  The new showCancel value
     */
    public void setShowCancel( boolean b ) {
        showCancel = b;
    }

    /**
     * Gets the showCancel attribute of the ChoiceInput object
     *
     * @return   The showCancel value
     */
    public boolean getShowCancel() {
        return showCancel;
    }

    public static ChoiceInput createYesNoChoice( String question ) {
        List list = new ArrayList();
        YES_CHOICE.setValue( bundle.getString( "yes" ) );
        NO_CHOICE.setValue( bundle.getString( "no" ) );
        list.add( YES_CHOICE );
        list.add( NO_CHOICE );
        ChoiceInput ci = new ChoiceInput( list, false );
        ci.setIntro( question );
        return ci;
    }

    public static ChoiceInput createYesNoCancelChoice( String question ) {
        List list = new ArrayList();
        list.add( bundle.getString( "yes" ) );
        list.add( bundle.getString( "no" ) );
        ChoiceInput ci = new ChoiceInput( list, true );
        ci.setIntro( question );
        return ci;
    }
}

