/*
Copyright (c) Dale Anson, 2004

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice,
 this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.
 3. The name of the author may not be used to endorse or promote products
 derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package ise.library.ascii;

import ise.library.Log;


/**
 * Produces colored output on terminals that support color, produces lots of
 * garbage on those terminals that don't.  The default color can be set in a
 * system property named "color", allowed values are "black", "red", "white",
 * "yellow", "blue", "green", and "magenta", any other value will display as
 * "cyan".  Further, a system property named "show_color" must have a value of
 * "yes" for the color codes to be included in the return value from
 * <code>toString</code>.
 *
 * @author Dale Anson
 *
 * @version   $Revision: 10 $
 */
public class ColoredString implements CharSequence {

    private StringBuffer contents = new StringBuffer();
    private StringBuffer coloredContents = new StringBuffer();

    /**
     * Constructor for ColoredString, default color is system default.
     *
     * @param cs The string that wants to be colored.
     */
    public ColoredString( CharSequence cs ) {
        if ( cs == null )
            cs = "";
        append( cs );
    }

    /**
     * Constructor for ColoredString
     *
     * @param cs The string that wants to be colored.
     * @param color The color.
     */
    public ColoredString( CharSequence cs, AnsiColor color ) {
        if ( cs == null )
            cs = "";
        append( cs, color );
    }

    public ColoredString append( CharSequence cs ) {
        contents.append( cs.toString() );
        coloredContents.append( cs.toString() );
        return this;
    }

    public ColoredString append( CharSequence cs, AnsiColor color ) {
        contents.append( cs.toString() );
        coloredContents.append( color ).append( cs.toString() ).append( AnsiColor.END_COLOR );
        return this;
    }

    /**
     * Returns the character at the specified index.
     *
     * @param index The index
     * @return      The character
     */
    public char charAt( int index ) {
        return contents.charAt( index );
    }

    /**
     * Returns the length of the sequence, could be 0.
     *
     * @return   The length of the sequence.
     */
    public int length() {
        return contents.length();
    }

    /**
     * Utility method to determine the length of a colorized string.
     * @param s a string to determine the length of
     * @return the length of the given string minus the length of any ansi color
     * control codes.
     */
    public static int length( CharSequence s ) {
        if (s == null)
            return 0;
        String t = new String( s.toString() );
        t = t.replaceAll( AnsiColor.ESCAPED_END_COLOR, "" );
        t = t.replaceAll( AnsiColor.CYAN.toEscapedString(), "" );
        t = t.replaceAll( AnsiColor.BLACK.toEscapedString(), "" );
        t = t.replaceAll( AnsiColor.RED.toEscapedString(), "" );
        t = t.replaceAll( AnsiColor.GREEN.toEscapedString(), "" );
        t = t.replaceAll( AnsiColor.YELLOW.toEscapedString(), "" );
        t = t.replaceAll( AnsiColor.BLUE.toEscapedString(), "" );
        t = t.replaceAll( AnsiColor.MAGENTA.toEscapedString(), "" );
        t = t.replaceAll( AnsiColor.WHITE.toEscapedString(), "" );
        t = t.replaceAll( AnsiColor.ERROR_COLOR.toEscapedString(), "" );
        t = t.replaceAll( AnsiColor.WARN_COLOR.toEscapedString(), "" );
        return t.length();
    }

    /**
     * Returns that portion of the sequence between the given indices.
     *
     * @param start The start index, inclusive.
     * @param end   The end index, exclusive.
     * @return      The subsequence.
     */
    public CharSequence subSequence( int start, int end ) {
        return contents.subSequence( start, end );
    }

    /**
     * The string embedded in an ansi color string.  This is what to send to the
     * terminal.
     *
     * @return   The colored string.
     */
    public String toString() {
        if ( "yes".equals( System.getProperty( "show_color" ) ) ) {
            return coloredContents.toString();
        }
        return contents.toString();
    }
}
