/*
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2003 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "Ant" and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */
package launcher;
import java.io.File;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.StringTokenizer;

import ise.dbconsole.DbConsole;

/**
 * This is a launcher for Ant. 
 * danson -- I've modified this to be a more generic
 * launcher. It's easier for an end-user to type 'java -jar myapp.jar' than it
 * is to figure out a classpath and so on. Certain jars may not be known at jar
 * time, so can't be added to the classpath set in the jar manifest. A good
 * example is a database driver, which might change after deployment of the app.
 * This class now assumes that the app name will be set, and that a directory
 * named .appname might be in user.home. It will load jar and zip files from
 * there as well as in the directory that the app is started from. A -lib
 * parameter may be used also to specify a location or locations to look for
 * library files.
 *
 * @author    Conor MacNeill
 * @version   $Revision: 7 $
 * @since     Ant 1.6
 */
public class Launcher {

    /** when setting up a new app, change this. */
    public final static String APP_NAME = "dbconsole";

    /** The location of a per-user library directory */
    public final static String USER_LIBDIR = "." + APP_NAME + "/lib";

    /** The startup class that is to be run, when setting up a new app change this. */
    public final static String MAIN_CLASS = "ise.dbconsole.DbConsole";

    /**
     * Entry point for starting command line
     *
     * @param args  commandline arguments
     */
    public static void main(String[] args) {
        try {
            Launcher launcher = new Launcher();
            launcher.run(args);
        }
        catch (LaunchException e) {
            System.err.println(e.getMessage());
        }
        catch (Throwable t) {
            t.printStackTrace();
        }
    }


    /**
     * Run the launcher to launch the main class
     *
     * @param args                       the command line arguments
     * @exception MalformedURLException  if the URLs required for the
     *      classloader cannot be created.
     * @exception LaunchException        Description of Exception
     */
    private void run(String[] args) throws LaunchException, MalformedURLException {

        File sourceJar = Locator.getClassSource(getClass());
        File jarDir = sourceJar.getParentFile();

        List libPaths = new ArrayList();
        List argList = new ArrayList();
        String[] newArgs;

        for (int i = 0; i < args.length; ++i) {
            if (args[i].equals("-lib")) {
                if (i == args.length - 1) {
                    throw new LaunchException("The -lib argument must "
                             + "be followed by a library location");
                }
                libPaths.add(args[++i]);
            }
            else {
                argList.add(args[i]);
            }
        }

        if (libPaths.size() == 0) {
            newArgs = args;
        }
        else {
            newArgs = (String[]) argList.toArray(new String[0]);
        }

        List libPathURLs = new ArrayList();
        for (Iterator i = libPaths.iterator(); i.hasNext(); ) {
            String libPath = (String) i.next();
            StringTokenizer myTokenizer
                     = new StringTokenizer(libPath, System.getProperty("path.separator"));
            while (myTokenizer.hasMoreElements()) {
                File element = new File(myTokenizer.nextToken());
                if (element.isDirectory()) {
                    // add any jars in the directory
                    URL[] dirURLs = Locator.getLocationURLs(element);
                    for (int j = 0; j < dirURLs.length; ++j) {
                        libPathURLs.add(dirURLs[j]);
                    }
                }
                libPathURLs.add(element.toURL());
            }
        }

        URL[] libJars = (URL[]) libPathURLs.toArray(new URL[0]);

        URL[] systemJars = Locator.getLocationURLs(jarDir);

        File userLibDir
                 = new File(System.getProperty("user.home"), USER_LIBDIR);
        URL[] userJars = Locator.getLocationURLs(userLibDir);

        int numJars = libJars.length + userJars.length + systemJars.length;
        URL[] jars = new URL[numJars];
        System.arraycopy(libJars, 0, jars, 0, libJars.length);
        System.arraycopy(userJars, 0, jars, libJars.length, userJars.length);
        System.arraycopy(systemJars, 0, jars, userJars.length + libJars.length,
                systemJars.length);

        // now update the class.path property
        StringBuffer baseClassPath
                 = new StringBuffer(System.getProperty("java.class.path"));
        if (baseClassPath.charAt(baseClassPath.length() - 1)
                 == File.pathSeparatorChar) {
            baseClassPath.setLength(baseClassPath.length() - 1);
        }

        for (int i = 0; i < jars.length; ++i) {
            baseClassPath.append(File.pathSeparatorChar);
            baseClassPath.append(Locator.fromURI(jars[i].toString()));
        }

        System.setProperty("java.class.path", baseClassPath.toString());

        URLClassLoader loader = new URLClassLoader(jars);
        Thread.currentThread().setContextClassLoader(loader);
        Class mainClass = null;
        try {
            mainClass = loader.loadClass(MAIN_CLASS);
            mainClass.newInstance();
        }
        catch (Throwable t) {
            // The class may be a singleton.  The singleton pattern says there
            // will be exactly one method that returns an instance of the class
            // we're trying to load, so look through the methods in the main
            // class for such a method.  Further, the method that produces the
            // singleton is generally public and static, so check for those
            // modifiers.
            try {
                Method[] methods = mainClass.getDeclaredMethods();
                ArrayList candidates = new ArrayList();
                for (int i = 0; i < methods.length; i++) {
                    Method m = methods[i];
                    Class returnType = m.getReturnType();
                    if (returnType.getName().equals(mainClass.getName())) {
                        int modifiers = m.getModifiers();
                        if (Modifier.isPublic(modifiers) && Modifier.isStatic(modifiers))
                            candidates.add(m);
                    }
                }
                if (candidates.size() == 1) {
                    Method m = (Method)candidates.get(0);
                    m.invoke(mainClass, null);
                    return;
                }
            }
            catch (Exception e) {
                e.printStackTrace();
            }
            System.err.println("Unable to load main class: " + MAIN_CLASS);
        }
    }
}

